"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_promise_1 = require("child-process-promise");
const fs_1 = __importDefault(require("mz/fs"));
const path_1 = require("path");
const rimraf_1 = __importDefault(require("rimraf"));
const util_1 = require("util");
class DependencyManager {
    constructor(rootPath, logger, gitHostWhitelist) {
        this.rootPath = rootPath;
        this.logger = logger;
        this.gitHostWhitelist = gitHostWhitelist;
        this.rimrafAysnc = util_1.promisify(rimraf_1.default);
    }
    installDependency() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.runNpm();
            }
            catch (e) {
                // console.debug(e)
            }
        });
    }
    runNpm() {
        return __awaiter(this, void 0, void 0, function* () {
            const env = Object.create(process.env);
            env.TERM = 'dumb';
            const cwd = this.rootPath;
            const yarnScript = require.resolve('yarn/bin/yarn.js');
            // if (existsSync(resolve(cwd, 'package-lock.json')) && !existsSync(resolve(cwd, 'yarn.lock'))) {
            //  TODO, use bundle npm?
            // cmd = 'npm';
            // }
            // TODO try to find package.json in lower level of directory
            this.logger.info('Running yarn to install deps..');
            if (!fs_1.default.existsSync(path_1.resolve(cwd, 'package.json'))) {
                return;
            }
            // const Spawn
            try {
                const promise = child_process_promise_1.spawn(process.execPath, [
                    yarnScript,
                    'install',
                    '--json',
                    '--ignore-scripts',
                    '--no-progress',
                    '--non-interactive',
                    '--ignore-engines',
                    '--pure-lockfile',
                    '--link-duplicates',
                ], {
                    env,
                    cwd,
                });
                const MAX_YARN_OUTPUT = 200;
                let totalLength = 0;
                promise.childProcess.stdout.on('data', (data) => {
                    if (totalLength < MAX_YARN_OUTPUT) {
                        const msg = data.toString();
                        this.logger.info('[yarn]', msg);
                        totalLength += msg.length;
                    }
                });
                promise.childProcess.stderr.on('data', (data) => {
                    if (totalLength < MAX_YARN_OUTPUT) {
                        const msg = data.toString();
                        this.logger.info('[yarn]', msg);
                        totalLength += msg.length;
                    }
                });
                yield promise;
            }
            catch (e) {
                this.logger.error('Can\'t launch yarn to download dependencies' + e.toString());
            }
            this.logger.info('Filtering un whitelisted packages');
            yield this.deletePackageRecursively(path_1.resolve(this.rootPath, 'node_modules'));
        });
    }
    deletePackageRecursively(rootPath) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const names = yield fs_1.default.readdir(rootPath);
                for (const name of names) {
                    const childPath = path_1.resolve(rootPath, name);
                    const stat = yield fs_1.default.stat(childPath);
                    if (stat.isDirectory()) {
                        const packageFile = path_1.resolve(childPath, 'package.json');
                        if (fs_1.default.existsSync(packageFile)) {
                            const fileContent = yield fs_1.default.readFile(packageFile, 'utf-8');
                            try {
                                const json = JSON.parse(fileContent);
                                this.logger.log('Checking package:' + json.name);
                                if (json.repository) {
                                    const url = new URL(json.repository.url);
                                    let allowed = false;
                                    for (const host of this.gitHostWhitelist) {
                                        if (url.hostname === host || url.hostname.endsWith('.' + host)) {
                                            allowed = true;
                                            break;
                                        }
                                    }
                                    if (!allowed) {
                                        this.logger.info('Deleting package that is not whitelisted: ' + json.name + '(' + url.href + ')');
                                        try {
                                            yield this.rimrafAysnc(childPath, {});
                                            continue;
                                        }
                                        catch (e) {
                                            this.logger.error('Deleting repo error: ' + e.toString());
                                        }
                                    }
                                }
                            }
                            catch (e) {
                                this.logger.log('Deleting repo error: ' + e);
                            }
                            const nodeModulesDir = path_1.resolve(childPath, 'node_modules');
                            if (fs_1.default.existsSync(nodeModulesDir)) {
                                yield this.deletePackageRecursively(nodeModulesDir);
                            }
                        }
                        else {
                            yield this.deletePackageRecursively(childPath);
                        }
                    }
                }
            }
            catch (e) {
                this.logger.log(e);
            }
        });
    }
}
exports.DependencyManager = DependencyManager;
//# sourceMappingURL=dependency-manager.js.map