"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const document_symbol_1 = require("@elastic/typescript-language-server/lib/document-symbol");
const lsp_server_1 = require("@elastic/typescript-language-server/lib/lsp-server");
const protocol_translation_1 = require("@elastic/typescript-language-server/lib/protocol-translation");
const mz_1 = require("mz");
const fs_1 = require("mz/fs");
const source_map_1 = require("source-map");
const path = __importStar(require("path"));
const url_1 = require("url");
const vscode_languageserver_1 = require("vscode-languageserver");
const dependencies_1 = require("./dependencies");
const dependency_manager_1 = require("./dependency-manager");
const NODE_MODULES = path.sep + 'node_modules' + path.sep;
const EMPTY_FULL = { symbols: [], references: [] };
const TYPESCRIPT_DIR_URI = url_1.pathToFileURL(path.resolve(require.resolve('typescript'), '..', '..', 'package.json') + '/');
const TYPESCRIPT_VERSION = JSON.parse(mz_1.fs.readFileSync(path.resolve(require.resolve('typescript'), '..', '..', 'package.json'), 'utf-8')).version;
class ExtendedLspServer extends lsp_server_1.LspServer {
    documentSymbol(params) {
        this.ensureDocumentOpen(params.textDocument.uri);
        return super.documentSymbol(params);
    }
    hover(params) {
        return __awaiter(this, void 0, void 0, function* () {
            this.ensureDocumentOpen(params.textDocument.uri);
            const file = protocol_translation_1.uriToPath(params.textDocument.uri);
            this.logger.log('hover', params, file);
            if (!file) {
                return { contents: [] };
            }
            const result = yield this.interuptDiagnostics(() => this.getQuickInfo(file, params.position));
            if (!result || !result.body) {
                return { contents: [] };
            }
            const range = protocol_translation_1.asRange(result.body);
            const contents = [
                { language: 'typescript', value: this.replaceWorkspaceInString(result.body.displayString) },
            ];
            // TODO: security filtering for documentation
            const tags = protocol_translation_1.asTagsDocumentation(result.body.tags);
            contents.push(result.body.documentation + (tags ? '\n\n' + tags : ''));
            return {
                contents,
                range,
            };
        });
    }
    references(params) {
        const _super = Object.create(null, {
            references: { get: () => super.references }
        });
        return __awaiter(this, void 0, void 0, function* () {
            this.ensureDocumentOpen(params.textDocument.uri);
            const rawResult = yield _super.references.call(this, params);
            return rawResult.filter((location) => {
                return location.uri.indexOf(NODE_MODULES) === -1;
            });
        });
    }
    initialize(params) {
        const _super = Object.create(null, {
            initialize: { get: () => super.initialize }
        });
        return __awaiter(this, void 0, void 0, function* () {
            const dependencyManager = new dependency_manager_1.DependencyManager(params.rootPath || protocol_translation_1.uriToPath(params.rootUri), this.logger, params.initializationOptions.gitHostWhitelist);
            if (params.initializationOptions.installNodeDependency) {
                yield dependencyManager.installDependency();
            }
            const result = yield _super.initialize.call(this, params);
            // // @ts-ignore
            // const args: protocol.SetCompilerOptionsForInferredProjectsArgs = {
            //   options: {
            //     module: protocol.ModuleKind.CommonJS,
            //     allowNonTsExtensions: false,
            //     lib: [
            //       'esnext',
            //       'dom',
            //     ],
            //   },
            // };
            // // @ts-ignore
            // const r = await this.tspClient.request(CommandTypes.CompilerOptionsForInferredProjects, args);
            // this.logger.error(r);
            return result;
        });
    }
    edefinition(params) {
        return __awaiter(this, void 0, void 0, function* () {
            this.ensureDocumentOpen(params.textDocument.uri);
            const definition = yield this.definition(params);
            if (definition === null) {
                return null;
            }
            if (definition instanceof Array) {
                // TODO do we want to deal with the case more defintions are presented?
                return this.convertLocation(definition[0]);
            }
            return this.convertLocation(definition);
        });
    }
    full(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const uri = params.textDocument.uri;
            if (uri.endsWith('bundle.js') || uri.endsWith('.min.js')) {
                return EMPTY_FULL;
            }
            this.ensureDocumentOpen(params.textDocument.uri);
            // const req = {
            //   needFileNameList: false,
            //   file: uriToPath(uri),
            // };
            // const projectInfoResp = await this.tspClient.request(CommandTypes.ProjectInfo, req);
            // this.logger.error(projectInfoResp);
            const file = protocol_translation_1.uriToPath(params.textDocument.uri);
            this.logger.log('full', params, file);
            if (!file) {
                return EMPTY_FULL;
            }
            const response = yield this.tspClient.request("navtree" /* NavTree */, {
                file,
            });
            const tree = response.body;
            if (!tree || !tree.childItems) {
                return EMPTY_FULL;
            }
            // if (this.supportHierarchicalDocumentSymbol) {
            //     const symbols: lsp.DocumentSymbol[] = [];
            //     for (const item of tree.childItems) {
            //         collectDocumentSymbols(item, symbols);
            //     }
            //     return symbols;
            // }
            const symbols = [];
            // return EMPTY_FULL;
            for (const item of tree.childItems) {
                document_symbol_1.collectSymbolInformations(params.textDocument.uri, item, symbols);
            }
            this.didCloseTextDocument({ textDocument: { uri: params.textDocument.uri } });
            const filterSymbols = symbols.filter((s) => s.name.endsWith(' callback'));
            const detailSymbols = yield Promise.all(filterSymbols.map((s) => this.toDetailSymbolInformation(s)));
            return { symbols: detailSymbols, references: [] };
        });
    }
    didOpenTextDocument(params) {
        const file = protocol_translation_1.uriToPath(params.textDocument.uri);
        this.logger.log('onDidOpenTextDocument', params, file);
        if (!file) {
            return;
        }
        if (this.documents.open(file, params.textDocument)) {
            this.tspClient.notify("open" /* Open */, {
                file,
                scriptKindName: this.getScriptKindName(params.textDocument.languageId),
                projectRootPath: this.rootPath(),
            });
        }
    }
    exit() {
        process.exit();
    }
    // private cleanContainerName(name: string): string {
    //   return name.split('"').join('').split('\\').join('.').split('/').join('.');
    // }
    findTsserverPath() {
        return path.resolve(require.resolve('typescript'), '..', 'tsserver.js');
    }
    toDetailSymbolInformation(symbol) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = new url_1.URL(symbol.location.uri);
            const [, packageJson] = yield dependencies_1.findClosestPackageJson(url, url_1.pathToFileURL(this.rootPath())); // enough param?
            const [repoUri] = getRepoUri(packageJson);
            const packageLocator = {
                name: packageJson.name,
                repoUri,
                version: '',
            };
            return {
                symbolInformation: symbol,
                qname: getQnameBySymbolInformation(symbol),
                package: packageLocator,
            };
        });
    }
    ensureDocumentOpen(uri) {
        // TODO check if languageId does matter
        this.didOpenTextDocument({ textDocument: { uri, languageId: '', text: '', version: 0 } });
    }
    replaceWorkspaceInString(str) {
        const withoutRoot = str.replace(this.rootPath(), '/'); // rootPath has tailing /
        const res = withoutRoot.split(NODE_MODULES);
        return res[res.length - 1];
    }
    convertLocation(location) {
        return __awaiter(this, void 0, void 0, function* () {
            const url = new url_1.URL(location.uri);
            // TODO, we may not need this block, find a repo to test it
            if (url.href.startsWith(TYPESCRIPT_DIR_URI.href)) {
                const relativeFilePath = url.href.substr(url.href.indexOf(TYPESCRIPT_DIR_URI.href) - 1);
                // Can we do sourceMap?
                const typescriptUrl = `git://github.com/Microsoft/TypeScript/blob/${TYPESCRIPT_VERSION}/${relativeFilePath}`;
                return {
                    location: { uri: typescriptUrl, range: location.range },
                };
            }
            if (location.uri.includes(NODE_MODULES)) {
                const moduleUrl = new url_1.URL(location.uri.substr(0, location.uri.indexOf(NODE_MODULES) + NODE_MODULES.length));
                const [, packageJson] = yield dependencies_1.findClosestPackageJson(url, moduleUrl); // enough param?
                // if (!packageJson.repository)
                const [repoUri, subdir] = getRepoUri(packageJson);
                // const npmConfig = configuration['typescript.npmrc'] || {}
                const gitVersion = 'master';
                // If there is only single package.json, maybe just use tag?
                try {
                    // const packageMeta = await fetchPackageMeta(packageJson.name, packageJson.version, {});
                    // if (packageMeta.gitHead) {
                    //   gitVersion = packageMeta.gitHead;
                    // }
                }
                catch (e) {
                    // Keep using 'master'
                }
                let mappedUri;
                let mappedRange;
                try {
                    const sourceMapUri = new url_1.URL(url.href + '.map');
                    const sourceMapPath = url_1.fileURLToPath(sourceMapUri);
                    const sourceMap = yield fs_1.readFile(sourceMapPath, 'utf-8');
                    const consumer = yield new source_map_1.SourceMapConsumer(sourceMap, sourceMapUri.href);
                    let mappedStart;
                    let mappedEnd;
                    try {
                        mappedStart = consumer.originalPositionFor({
                            line: location.range.start.line + 1,
                            column: location.range.start.character,
                        });
                        mappedEnd = consumer.originalPositionFor({
                            line: location.range.end.line + 1,
                            column: location.range.end.character,
                        });
                    }
                    finally {
                        consumer.destroy();
                    }
                    if (mappedStart.source === null ||
                        mappedStart.line === null ||
                        mappedStart.column === null ||
                        mappedEnd.line === null ||
                        mappedEnd.column === null) {
                        // Error
                        mappedUri = url;
                        mappedRange = {
                            start: {
                                line: 0,
                                character: 0,
                            },
                            end: {
                                line: 0,
                                character: 0,
                            },
                        };
                    }
                    else {
                        mappedUri = new url_1.URL(mappedStart.source);
                        // if (!mappedUri.href.startsWith(tempDirUri.href)) {
                        //   throw new Error(
                        //     `Mapped source URI ${mappedUri} is not under root URI ${fileRootUri} and not in automatic typings`
                        //   )
                        // }
                        mappedRange = {
                            start: {
                                line: mappedStart.line - 1,
                                character: mappedStart.column,
                            },
                            end: {
                                line: mappedEnd.line - 1,
                                character: mappedEnd.column,
                            },
                        };
                    }
                }
                catch (e) {
                    mappedUri = url;
                    mappedRange = location.range;
                }
                // TODO this should be simplified
                const depRootDir = dependencies_1.resolveDependencyRootDir(url_1.fileURLToPath(url));
                const mappedPackageRelativeFilePath = path.posix.relative(depRootDir, url_1.fileURLToPath(mappedUri));
                const mappedRepoRelativeFilePath = path.posix.join(subdir, mappedPackageRelativeFilePath);
                const symbolLocation = {
                    uri: repoUri + '/blob/' + gitVersion + '/' + mappedRepoRelativeFilePath,
                    range: mappedRange,
                };
                return {
                    // TODO add qname
                    location: symbolLocation,
                    package: {
                        name: packageJson.name,
                        repoUri,
                        version: gitVersion,
                    },
                };
            }
            else {
                return {
                    location,
                };
            }
        });
    }
}
exports.ExtendedLspServer = ExtendedLspServer;
function getRepoUri(packageJson) {
    let cloneUrl = dependencies_1.cloneUrlFromPackageMeta(packageJson);
    let subdir = '';
    // Handle GitHub tree URLs
    const treeMatch = cloneUrl.match(/^(?:https?:\/\/)?(?:www\.)?github.com\/[^\/]+\/[^\/]+\/tree\/[^\/]+\/(.+)$/);
    if (treeMatch) {
        subdir = treeMatch[1];
        cloneUrl = cloneUrl.replace(/(\/tree\/[^\/]+)\/.+/, '$1');
    }
    if (typeof packageJson.repository === 'object' && packageJson.repository.directory) {
        subdir = packageJson.repository.directory;
    }
    else if (packageJson.name.startsWith('@types/')) {
        // Special-case DefinitelyTyped
        subdir = packageJson.name.substr(1);
    }
    let repoUri = cloneUrl.replace('https://', 'git://');
    if (repoUri.endsWith('.git')) {
        repoUri = repoUri.substr(0, repoUri.length - 4);
    }
    return [repoUri, subdir];
}
function getQnameBySymbolInformation(info) {
    let prefix = '';
    // if (packageLocator && packageLocator.name && packageLocator.name !== '') {
    //     prefix += packageLocator.name + '.'
    // } else {
    //     prefix = 'unknown'
    // }
    // const fileName = this.getFileName(info.location.uri);
    // const simpleName = this.getSimpleFileName(fileName)
    // if (info.location.uri !== '') {
    //     prefix += simpleName + '.'
    // }
    if (info.kind !== vscode_languageserver_1.SymbolKind.Field) {
        if (info.containerName && info.containerName !== '') {
            prefix += info.containerName + '.';
        }
    }
    return prefix + info.name;
}
//
// function collectSymbolInformations(
//   uri: string, current: tsp.NavigationTree, symbols: lsp.SymbolInformation[], containerName?: string): boolean {
//   let shouldInclude = shouldIncludeEntry(current);
//   const name = current.text;
//   for (const span of current.spans) {
//     const range = asRange(span);
//     const children: lsp.SymbolInformation[] = [];
//     if (current.childItems) {
//       for (const child of current.childItems) {
//         // if (child.spans.some((span2) => !!Range.intersection(range, asRange(span2)))) {
//         const includedChild = collectSymbolInformations(uri, child, children, name);
//         shouldInclude = shouldInclude || includedChild;
//         // }
//       }
//     }
//     if (shouldInclude) {
//       symbols.push({
//         name,
//         kind: toSymbolKind(current.kind),
//         location: {
//           uri,
//           range,
//         },
//         containerName,
//       });
//       symbols.push(...children);
//     }
//     break;
//   }
//
//   return shouldInclude;
// }
//# sourceMappingURL=extended-lsp-server.js.map